#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Nov 14 00:21:25 2021

@author: li
"""

from torch.optim import lr_scheduler
from warmup_scheduler import GradualWarmupScheduler
import torch
from torch.utils.data import Dataset
import torchvision.transforms as transforms
from PIL import Image
import os


    
    


def local_dataset(train_path,val_path,batch_size,data_type = 'png',
                        train_transform = None,val_transform = None,labellist = {},
                        num_workers=0):
    
    if train_transform == None:
            train_transform=transforms.Compose(
                [
                    # transforms.RandomRotation((60)),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                    ),
                ]
            )
    if val_transform == None:
            val_transform=transforms.Compose(
                [
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                    ),
                ]
            )
            
    train_set=MyDateset(
                    data_dir = train_path,
                    transform = train_transform,
                    data_type = data_type,
                    labellist = labellist,
                    )
    
    labellist = train_set.labellist
    
    train_loader = torch.utils.data.DataLoader(
        train_set,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers, 
        pin_memory=True, 
    )
    
    val_set = MyDateset(
                data_dir = val_path,
                transform = val_transform,
                data_type = data_type,
                labellist = labellist,
                )
    
    val_loader = torch.utils.data.DataLoader(
        val_set,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers, 
        pin_memory=True, 
    )
    
    return train_loader,val_loader,labellist

def local_dataset_test(test_path,data_type,batch_size,test_transform = None,labellist = {},
                       num_workers=0):

    if test_transform == None:
            test_transform=transforms.Compose(
                [
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                    ),
                ]
            )
    
    test_set = MyDateset(
                data_dir = test_path,
                transform = test_transform,
                data_type = data_type,
                labellist = labellist,
                )
    
    
    test_loader = torch.utils.data.DataLoader(
        test_set,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers, 
        pin_memory=True, 
    )
    
    return test_loader

class MyDateset(Dataset):
    def __init__(self,data_dir,transform=None,data_type='.JPEG',labellist={}):
        self.labellist = labellist
        self.count     = len(self.labellist)
        self.data_type = data_type
        self.data_info = self.get_img_info(data_dir)
        self.transform = transform
        
        
    def __getitem__(self, index):
        path_img, label = self.data_info[index] 
        img = Image.open(path_img).convert('RGB')
        if self.transform is not None:
            img = self.transform(img)
        return img, label     
        
    def __len__(self):
        return len(self.data_info)
    
    def get_img_info(self,data_dir): 
        data_info = list()
        for root, dirs, _ in os.walk(data_dir):
            for sub_dir in dirs:  
                if sub_dir not in self.labellist : 
                    self.labellist[sub_dir] = self.count
                    self.count+=1
                img_list = os.listdir(os.path.join(root, sub_dir))
                img_list = list(filter(lambda x: x.endswith(self.data_type), img_list))
                
                for i in range(len(img_list)):
                    img_name = img_list[i]
                    path_img = os.path.join(root, sub_dir, img_name) 
                    label = self.labellist[sub_dir]
                    data_info.append((path_img, label))   
        return data_info

def test(net,test_loader,device,criterion = None):
    
    correct = total = test_loss = 0
    net.eval()
    net = net.to(device=device)
    if criterion:
        with torch.no_grad():
            for idx, (inputs, targets) in enumerate(test_loader):
                if torch.cuda.is_available() :
                    inputs = inputs.cuda(non_blocking=True)
                    targets = targets.cuda(non_blocking=True)
                else :
                    inputs, targets = inputs.to(device), targets.to(device)
                    
                outputs = net(inputs)
                loss = criterion(outputs, targets)
                test_loss += loss.item()
                test_idx = (idx + 1)
                total += targets.size(0)
                correct += torch.eq(outputs.argmax(dim=1), targets).sum().item()
                # if (idx+1)%5 == 0:
                #     break
            test_loss /= test_idx 
            test_acc = correct / total *100.0
            
        return test_loss, test_acc
        
    else:
        with torch.no_grad():
            for idx, (inputs, targets) in enumerate(test_loader):
                if torch.cuda.is_available() :
                    inputs = inputs.cuda(non_blocking=True)
                    targets = targets.cuda(non_blocking=True)
                else :
                    inputs, targets = inputs.to(device), targets.to(device)
                    
                outputs = net(inputs)
                total += targets.size(0)  
                correct += torch.eq(outputs.argmax(dim=1), targets).sum().item()
                # if (idx+1)%5 == 0:
                #     break
            test_acc = correct / total *100.0
        
        return test_acc

 
def val(net,val_loader,device,criterion):
    
    val_loss = total = correct = 0
    net.eval() 
    net = net.to(device=device)
    with torch.no_grad():
        for idx, (inputs, targets) in enumerate(val_loader):
            if torch.cuda.is_available() :
                inputs = inputs.cuda(non_blocking=True)
                targets = targets.cuda(non_blocking=True)
            else :
                inputs, targets = inputs.to(device), targets.to(device)
                
            outputs = net(inputs)
            loss = criterion(outputs, targets)
            val_loss += loss.item()
            val_idx = (idx + 1)
            total += targets.size(0)
            correct += torch.eq(outputs.argmax(dim=1), targets).sum().item()
            
    val_loss /= val_idx 
    val_acc = correct / total *100.0
    
    return val_loss, val_acc
    

def train_stepLR(net,EPOCHS,train_loader,val_loader,device,criterion,optimizer,SAVE_NAME,
               log='',best_acc=0):
    
    best_loss = 0
    scheduler_cos = lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0 = 5, T_mult = 2)
    scheduler = GradualWarmupScheduler(optimizer, multiplier=1, total_epoch=5, after_scheduler=scheduler_cos)
    
    optimizer.zero_grad()
    optimizer.step()
    for ep in range(1, EPOCHS + 1):
        
        print("\n   == Epoches: [{}/{}] ============================".format(ep,EPOCHS))
        if log:
            with open(log,"a+") as f: 
                print("\n   == Epoches: [{}/{}] ============================".format(ep,EPOCHS),file = f)
                
        train_loss =  correct = total = 0
        net.train() 
        net = net.to(device=device)
        scheduler.step(ep)
        
        for idx, (inputs, targets) in enumerate(train_loader):
            if torch.cuda.is_available() :
                inputs = inputs.cuda(non_blocking=True)
                targets = targets.cuda(non_blocking=True)
            else :
                inputs, targets = inputs.to(device), targets.to(device)
                
            outputs = net(inputs)
            loss = criterion(outputs, targets)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            train_loss += loss.item()
            total += targets.size(0)
            correct += torch.eq(outputs.argmax(dim=1), targets).sum().item()
            
            if (idx + 1) % 100 == 0 or (idx + 1) == len(train_loader):
                
                print(
                    "   == step: [{:3}/{}] | loss: {:.3f} | acc: {:6.3f}%".format(
                        idx + 1,
                        len(train_loader),
                        train_loss / (idx + 1),
                        100.0 * correct / total,
                    ))
                if log:
                    with open(log,"a+") as f:
                        print(
                            "   == step: [{:3}/{}] | loss: {:.3f} | acc: {:6.3f}%".format(
                                idx + 1,
                                len(train_loader),
                                train_loss / (idx + 1),
                                100.0 * correct / total,
                            ),file = f)
                
        
        val_loss, val_acc = val(net,val_loader,device,criterion)
        
        print('   val_loss:',val_loss,'|  val_acc:',val_acc)
        if log:
            with open(log,"a+") as f: 
                print('   val_loss:',val_loss,'|  val_acc:',val_acc,file = f)
        
        if val_acc >= best_acc :  
            best_acc = val_acc
            best_loss = val_loss
            torch.save(net.state_dict(),SAVE_NAME +'.pth')  
            print('   Parameter save in ',ep)

            if log:
                with open(log,"a+") as f: 
                    print('   Parameter save in ',ep,file = f)
            # if val_acc >= 90:
            #     break
            
        print('   Now best acc is :',best_acc,'| best loss is :',best_loss)
        if log:
            with open(log,"a+") as f: 
                print('   Now best acc is :',best_acc,' | best loss is :',best_loss,file=f)
         
    return net